package mbd;
import java.util.Arrays;
import figure.DefaultMatrix;
import figure.Matrix;
import figure.Point3D;
import figure.Vector;
import figure.Vector3D;

//--------------------------------------------------------------------//
/**
 * ̃NX
 */
public abstract class Body implements Cloneable,LabeledData{

/** f[^x */
  public int label;

/**  */
  public double mass;

/** [gi[z */
  public double inertia[][];

/** ʒu */
  public Position p;

/** ʒu */
  public Position p0;

/**
 * ̂\z
 * @param label f[^x
 */
  public Body(int label){
    this.label=label;
    mass=1.0;
    inertia=new double[3][3];
    initInertia();
    p=new Position();
    p0=new Position();
  }

//--------------------------------------------------------------------//
/**
 * ̌`\_Ԃ
 * @return _i[z
 */
  public abstract Point3D[][] vertices();

//--------------------------------------------------------------------//
/**
 * f[^xԂ
 * @return f[^x
 */
  public int label(){
    return label;
  }

//--------------------------------------------------------------------//
/**
 * f[^xݒ肷
 * @param label f[^x
 */
  public void setLabel(int label){
    this.label=label;
  }

//--------------------------------------------------------------------//
/**
 * [g
 */
  protected void initInertia(){
    for(int i=0;i<3;i++){
      for(int j=0;j<3;j++){
      	inertia[i][j]=0.0;
      }
      inertia[i][i]=1.0;
    }
  }

//--------------------------------------------------------------------//
/**
 * ʁE[gݒ肷
 * @param mass 
 * @param ix <i>x</i> ̊[g
 * @param iy <i>y</i> ̊[g
 * @param iz <i>z</i> ̊[g
 */
  public void setMass(double mass,double ix,double iy,double iz){
    this.mass=mass;
    initInertia();
    inertia[0][0]=ix;
    inertia[1][1]=iy;
    inertia[2][2]=iz;
  }

//--------------------------------------------------------------------//
/**
 * ʃ}gbNXԂ
 * @return ʃ}gbNX
 */
  public Matrix massMatrix(){
    Matrix m=new DefaultMatrix(6);
    m.set(0,0,mass);
    m.set(1,1,mass);
    m.set(2,2,mass);
    for(int i=0;i<3;i++){
      for(int j=0;j<3;j++){
      	m.set(i+3,j+3,inertia[i][j]);
      }
    }
    return m;
  }

//--------------------------------------------------------------------//
/**
 * ʒuԂ
 * @return ʒu
 */
  public Position position(){
    return p;
  }

//--------------------------------------------------------------------//
/**
 * ʒuݒ肷
 */
  public void setInitPosition(){
    p0=(Position)p.clone();
  }

//--------------------------------------------------------------------//
/**
 * ʒuԂɖ߂
 */
  public void initPosition(){
    p=(Position)p0.clone();
  }

//--------------------------------------------------------------------//
/**
 * Ỏ͖׏dxNgԂ
 * @param p ʒu
 * @return ׏dxNg
 */
  public Vector forceVector(Position p){
    Vector3D v=new Vector3D();
    return forceVector(v,v,p);
  }

//--------------------------------------------------------------------//
/**
 * ׏dxNgԂ
 * @param f O
 * @param t OgN
 * @param p ʒu
 * @return ׏dxNg
 */
  public Vector forceVector(Vector3D f,Vector3D t,Position p){
    Vector v=new Vector(6,new double[]{f.x,f.y,f.z,t.x,t.y,t.z});
    Vector jw=new Vector(3);
    for(int i=0;i<3;i++){
      jw.set(i,inertia[i][0]*p.w.x+inertia[i][1]*p.w.y
      	      +inertia[i][2]*p.w.z);
    }
    jw=jw.mul(p.w.tilde());
    for(int i=0;i<3;i++){
      v.add(i+3,-jw.get(i));
    }
    return v;
  }

//--------------------------------------------------------------------//
/**
 * IuWFNg̃Rs[Ԃ
 * @return ̃CX^X̕
 */
  public Object clone(){
    try{
      Body body=(Body)super.clone();
      body.inertia=new double[3][3];
      for(int i=0;i<3;i++){
      	for(int j=0;j<3;j++){
      	  body.inertia[i][j]=inertia[i][j];
      	}
      }
      body.p=(Position)p.clone();
      return body;
    }
    catch(CloneNotSupportedException ex){
      throw new InternalError();
    }
  }

//--------------------------------------------------------------------//
/**
 * IuWFNg̃nbVR[hlԂ
 * @return IuWFNg̃nbVR[hl
 */
  public int hashCode(){
    long lm=Double.doubleToLongBits(mass);
    int bit=(int)(lm^(lm>>>32));
    for(int i=0;i<3;i++){
      for(int j=0;j<3;j++){
      	lm=Double.doubleToLongBits(inertia[i][j]);
      	bit=31*bit+(int)(lm^(lm>>>32));
      }
    }
    bit=31*bit+p.hashCode();
    return bit;
  }

//--------------------------------------------------------------------//
/**
 * IuWFNg̓
 * @param obj rΏۂ̃IuWFNg
 * @return rΏۂ̃IuWFNg <code>Body</code>
 *         ̃CX^Xœlꍇ true
 */
  public boolean equals(Object obj){
    if(obj instanceof Body){
      Body b=(Body)obj;
      return ((mass==b.mass)
      	   && Arrays.equals(inertia[0],b.inertia[0])
      	   && Arrays.equals(inertia[1],b.inertia[1])
      	   && Arrays.equals(inertia[2],b.inertia[2])
      	   && p.equals(b.p));
    }
    else{
      return false;
    }
  }
}
