package mbd;
import figure.DefaultMatrix;
import figure.Matrix;

//--------------------------------------------------------------------//
/**
 * IC[p[^
 */
public class EulerParameters implements Cloneable{

/** p[^ p0=cos/2 */
  public double p0;

/** p[^ p1=u<sub>x</sub>sin/2 */
  public double p1;

/** p[^ p2=u<sub>y</sub>sin/2 */
  public double p2;

/** p[^ p3=u<sub>z</sub>sin/2 */
  public double p3;

/**
 * IC[p[^\z
 */
  public EulerParameters(){
    this(1.0,0.0,0.0,0.0);
  }

/**
 * IC[p[^\z
 * @param p0 p[^ p0=cos/2
 * @param p1 p[^ p1=u<sub>x</sub>sin/2
 * @param p2 p[^ p2=u<sub>y</sub>sin/2
 * @param p3 p[^ p3=u<sub>z</sub>sin/2
 */
  public EulerParameters(double p0,double p1,double p2,double p3){
    set(p0,p1,p2,p3);
  }

//--------------------------------------------------------------------//
/**
 * IC[p[^ݒ肷
 * @param p0 p[^ p0=cos/2
 * @param p1 p[^ p1=u<sub>x</sub>sin/2
 * @param p2 p[^ p2=u<sub>y</sub>sin/2
 * @param p3 p[^ p3=u<sub>z</sub>sin/2
 */
  public void set(double p0,double p1,double p2,double p3){
    this.p0=p0;
    this.p1=p1;
    this.p2=p2;
    this.p3=p3;
  }

//--------------------------------------------------------------------//
/**
 * IC[p[^ɒl
 * @param dp0 p0 ɉl
 * @param dp1 p1 ɉl
 * @param dp2 p2 ɉl
 * @param dp3 p3 ɉl
 */
  public void add(double dp0,double dp1,double dp2,double dp3){
    p0+=dp0;
    p1+=dp1;
    p2+=dp2;
    p3+=dp3;
  }

//--------------------------------------------------------------------//
/**
 * ]sԂ
 * @return ]s
 */
  public Matrix rotateMatrix(){
    Matrix a=new DefaultMatrix(3);
    double p02=p0*p0;
    a.set(0,0,2.0*(p02+p1*p1)-1.0);
    a.set(0,1,2.0*(p1*p2-p0*p3));
    a.set(0,2,2.0*(p1*p3+p0*p2));
    a.set(1,0,2.0*(p1*p2+p0*p3));
    a.set(1,1,2.0*(p02+p2*p2)-1.0);
    a.set(1,2,2.0*(p2*p3-p0*p1));
    a.set(2,0,2.0*(p1*p3-p0*p2));
    a.set(2,1,2.0*(p2*p3+p0*p1));
    a.set(2,2,2.0*(p02+p3*p3)-1.0);
    return a;
  }

//--------------------------------------------------------------------//
/**
 * s E Ԃ
 * @return s E
 */
  public Matrix matrixE(){
    Matrix e=new DefaultMatrix(3,4);
    e.set(0,0,-p1);
    e.set(0,1,p0);
    e.set(0,2,-p3);
    e.set(0,3,p2);
    e.set(1,0,-p2);
    e.set(1,1,p3);
    e.set(1,2,p0);
    e.set(1,3,-p1);
    e.set(2,0,-p3);
    e.set(2,1,-p2);
    e.set(2,2,p1);
    e.set(2,3,p0);
    return e;
  }

//--------------------------------------------------------------------//
/**
 * s L Ԃ
 * @return s L
 */
  public Matrix matrixL(){
    Matrix l=new DefaultMatrix(3,4);
    l.set(0,0,-p1);
    l.set(0,1,p0);
    l.set(0,2,p3);
    l.set(0,3,-p2);
    l.set(1,0,-p2);
    l.set(1,1,-p3);
    l.set(1,2,p0);
    l.set(1,3,p1);
    l.set(2,0,-p3);
    l.set(2,1,p2);
    l.set(2,2,-p1);
    l.set(2,3,p0);
    return l;
  }

//--------------------------------------------------------------------//
/**
 * K
 */
  public void normalize(){
    double p=p0*p0+p1*p1+p2*p2+p3*p3;
    if(p==0.0) return;
    p=1.0/Math.sqrt(p);
    p0*=p;
    p1*=p;
    p2*=p;
    p3*=p;
  }

//--------------------------------------------------------------------//
/**
 * IuWFNg̃Rs[Ԃ
 * @return ̃CX^X̕
 */
  public Object clone(){
    try{
      return super.clone();
    }
    catch(CloneNotSupportedException ex){
      throw new InternalError();
    }
  }

//--------------------------------------------------------------------//
/**
 * IuWFNg̃nbVR[hlԂ
 * @return IuWFNg̃nbVR[hl
 */
  public int hashCode(){
    long l[]={Double.doubleToLongBits(p0),
      	      Double.doubleToLongBits(p1),
      	      Double.doubleToLongBits(p2),
      	      Double.doubleToLongBits(p3)};
    int bit=0;
    for(int i=0;i<l.length;i++){
      bit=31*bit+(int)(l[i]^(l[i]>>>32));
    }
    return bit;
  }

//--------------------------------------------------------------------//
/**
 * IuWFNg̓
 * @param obj rΏۂ̃IuWFNg
 * @return rΏۂ̃IuWFNg <code>EulerParameters</code>
 *         ̃CX^Xœlꍇ true
 */
  public boolean equals(Object obj){
    if(obj instanceof EulerParameters){
      EulerParameters ep=(EulerParameters)obj;
      return ((p0==ep.p0) && (p1==ep.p1) && (p2==ep.p2) && (p3==ep.p3));
    }
    else{
      return false;
    }
  }

//--------------------------------------------------------------------//
/**
 * f[^Ԃ
 * @return ̃IuWFNg̕\
 */
  public String toString(){
    return "EulerParameters\t"+p0+"\t"+p1+"\t"+p2+"\t"+p3;
  }
}
