package mbd;
import java.util.ArrayList;
import java.util.Collections;
import figure.Cube3D;
import figure.Matrix;
import figure.Point3D;
import figure.Tuple3D;
import figure.Vector;
import figure.Vector3D;
import figure.matrix.DefaultMatrixSolver;
import figure.matrix.MatrixSolveException;
import mbd.constraint.Spring;

//--------------------------------------------------------------------//
/**
 * ̌nf[^f
 */
public class MultiBodyModel{

/** Runge - Kutta @̌W 1 */
  public static final double C1[]={1.0/6.0,1.0/3.0,1.0/3.0,1.0/6.0};

/** Runge - Kutta @̌W 2 */
  public static final double C2[]={0.5,0.5,1.0};

/** x̔rIuWFNg */
  public static final LabelComparator
    LABEL_COMPARATOR=new LabelComparator();

/** S̔rIuWFNg */
  public static final ConstraintComparator
    CONST_COMPARATOR=new ConstraintComparator();

/** ̂i[郊Xg */
  protected ArrayList<Body> bodies;

/** Si[郊Xg */
  protected ArrayList<RigidConstraint> rigids;

/** _Si[郊Xg */
  protected ArrayList<Constraint> springs;

/** }gbNXExNg̃TCY */
  protected int size;

/** ʁES}gbNX */
  protected DefaultMatrixSolver matrix;

/** ׏dxNg */
  protected Vector vector;

/** Runge-Kutta @Xebṽf[^ */
  protected Position tempPos[][];

/** fE\ <code>Cube3D</code> */
  public Cube3D bounds;

/**  */
  protected Time time;

/** d͉xxNg */
  protected Vector3D g;

/**
 * ̌nf[^f\z
 */
  public MultiBodyModel(){
    bodies=new ArrayList<Body>();
    rigids=new ArrayList<RigidConstraint>();
    springs=new ArrayList<Constraint>();
    size=0;
    bounds=new Cube3D();
    time=new Time(1.0);
    g=new Vector3D(0.0,0.0,-9.80665);
  }

//--------------------------------------------------------------------//
/**
 * fEԂ
 * @return fE\
 */
  public Cube3D getBounds(){
    return bounds;
  }

//--------------------------------------------------------------------//
/**
 * ԂԂ
 * @return 
 */
  public Time getTime(){
    return time;
  }

//--------------------------------------------------------------------//
/**
 * ԍݕݒ肷
 * @param dt ԍݕ
 */
  public void setDt(double dt){
    time=new Time(dt);
  }

//--------------------------------------------------------------------//
/**
 * d͉xݒ肷
 * @param g d͉xxNg
 */
  public void setGravity(Vector3D g){
    this.g=g;
  }

//--------------------------------------------------------------------//
/**
 * ̂̐Ԃ
 * @return ̂̐
 */
  public int getBodyCount(){
    return bodies.size();
  }

//--------------------------------------------------------------------//
/**
 * ̂Ԃ
 * @param index ̂̃CfbNX
 * @return w肳ꂽʒu̕
 */
  public Body getBody(int index){
    return bodies.get(index);
  }

//--------------------------------------------------------------------//
/**
 * ̂ǉ
 * @param body 
 */
  public void addBody(Body body){
    bodies.add(body);
  }

//--------------------------------------------------------------------//
/**
 * S̐Ԃ
 * @return S̐
 */
  public int getConstraintCount(){
    return rigids.size();
  }

//--------------------------------------------------------------------//
/**
 * SԂ
 * @param index S̃CfbNX
 * @return w肳ꂽʒu̍S
 */
  public RigidConstraint getConstraint(int index){
    return rigids.get(index);
  }

//--------------------------------------------------------------------//
/**
 * Sǉ
 * @param constraint S
 */
  public void addConstraint(RigidConstraint constraint){
    rigids.add(constraint);
  }

//--------------------------------------------------------------------//
/**
 * _S̐Ԃ
 * @return S̐
 */
  public int getSpringCount(){
    return springs.size();
  }

//--------------------------------------------------------------------//
/**
 * _SԂ
 * @param index S̃CfbNX
 * @return w肳ꂽʒu̍S
 */
  public Constraint getSpring(int index){
    return springs.get(index);
  }

//--------------------------------------------------------------------//
/**
 * _Sǉ
 * @param spring S
 */
  public void addSpring(Spring spring){
    springs.add(spring);
  }

//--------------------------------------------------------------------//
/**
 * oEKë艻p[^ݒ肷
 * @param alpha oEKë艻p[^
 * @param beta oEKë艻p[^
 */
  public void setBaumgarte(double alpha,double beta){
    int constraintCount=getConstraintCount();
    for(int i=0;i<constraintCount;i++){
      getConstraint(i).setBaumgarte(alpha,beta);
    }
  }

//--------------------------------------------------------------------//
/**
 * f[^
 */
  public void clear(){
    bodies.clear();
    rigids.clear();
    springs.clear();
    size=0;
    bounds=new Cube3D();
  }

//--------------------------------------------------------------------//
/**
 * f
 */
  public void init(){
    bodies.trimToSize();
    rigids.trimToSize();
    springs.trimToSize();
    Collections.sort(bodies,LABEL_COMPARATOR);
    reNumbering();
    Collections.sort(rigids,CONST_COMPARATOR);
    Collections.sort(springs,CONST_COMPARATOR);
    setSize();
    setTempPosition();
    setGcPosition();
    setInitPosition();
    setBounds();
  }

//--------------------------------------------------------------------//
/**
 * ߓ_|C^ݒ肷
 */
  public void reNumbering(){
    LabelMap labelMap=new LabelMap();
    int bodyCount=getBodyCount();
    for(int i=0;i<bodyCount;i++){
      labelMap.put(getBody(i),i);
    }
    int constraintCount=getConstraintCount();
    for(int i=0;i<constraintCount;i++){
      labelMap.reset(getConstraint(i));
    }
    int springCount=getSpringCount();
    for(int i=0;i<springCount;i++){
      labelMap.reset(getSpring(i));
    }
  }

//--------------------------------------------------------------------//
/**
 * }gbNXExNg̃TCYݒ肷
 */
  protected void setSize(){
    size=6*getBodyCount();
    int constraintCount=getConstraintCount();
    for(int i=0;i<constraintCount;i++){
      size+=getConstraint(i).dof();
    }
  }

//--------------------------------------------------------------------//
/**
 * Runge-Kutta @Xebṽf[^ݒ肷
 */
  protected void setTempPosition(){
    int bodyCount=getBodyCount();
    tempPos=new Position[bodyCount][4];
    for(int i=0;i<bodyCount;i++){
      tempPos[i][0]=getBody(i).position();
    }
  }

//--------------------------------------------------------------------//
/**
 * dSʒuExݒ肷
 */
  protected void setGcPosition(){
    int constraintCount=getConstraintCount();
    for(int i=0;i<constraintCount;i++){
      RigidConstraint c=getConstraint(i);
      c.setGcPosition(getPosition(c.p1(),0),getPosition(c.p2(),0));
    }
  }

//--------------------------------------------------------------------//
/**
 * ʒuݒ肷
 */
  protected void setInitPosition(){
    time.init();
    int bodyCount=getBodyCount();
    for(int i=0;i<bodyCount;i++){
      getBody(i).setInitPosition();
    }
  }

//--------------------------------------------------------------------//
/**
 * fEݒ肷
 */
  protected void setBounds(){
    bounds.setBounds(0.0,0.0,0.0,0.0,0.0,0.0);
    int bodyCount=getBodyCount();
    if(bodyCount==0) return;
    for(int i=0;i<bodyCount;i++){
      Point3D pt[][]=getBody(i).vertices();
      for(int i1=0;i1<pt.length;i1++){
      	for(int j1=0;j1<pt[i1].length;j1++){
      	  if((i==0) && (i1==0) && (j1==0)){
      	    bounds.setLocation(pt[i1][j1]);
      	  }
      	  else{
      	    bounds.add(pt[i1][j1]);
      	  }
      	}
      }
    }
  }

//--------------------------------------------------------------------//
/**
 * ʒuԂɖ߂
 */
  public void initPosition(){
    time.init();
    int bodyCount=getBodyCount();
    for(int i=0;i<bodyCount;i++){
      getBody(i).initPosition();
    }
    setTempPosition();
    setGcPosition();
  }

//--------------------------------------------------------------------//
/**
 * ̃XebvvZ
 * @throws MatrixSolveException Ả܂ȂƂ
 */
  public void next() throws MatrixSolveException{
    time.next();
    double dt=time.dt();
    double tol=1.0e-10;
    for(int i=0;i<4;i++){
      setTempPosition(dt,i);
      createMatrixVector(i);
      setAcc(matrix.solveByLUNoPreserv(vector,tol),i);
    }
    update(dt);
  }

//--------------------------------------------------------------------//
/**
 * Runge-Kutta @Xebṽf[^ݒ肷
 * @param dt ԍݕ
 * @param step Runge - Kutta @̃Xebv
 */
  protected void setTempPosition(double dt,int step){
    if(step==0) return;
    double cdt=dt*C2[step-1];
    int bodyCount=getBodyCount();
    for(int i=0;i<bodyCount;i++){
      tempPos[i][step]=(Position)getBody(i).position().clone();
      tempPos[i][step].update(cdt,tempPos[i][step-1].v,
      	      	      	      tempPos[i][step-1].a,
      	      	      	      tempPos[i][step-1].eulerVel(),
      	      	      	      tempPos[i][step-1].dw);
    }
  }

//--------------------------------------------------------------------//
/**
 * ʁES}gbNXщ׏dxNg쐬
 * @param step Runge - Kutta @̃Xebv
 */
  protected void createMatrixVector(int step){
    matrix=new DefaultMatrixSolver(size);
    vector=new Vector(size);

    int ii=0;
    int bodyCount=getBodyCount();
    for(int i=0;i<bodyCount;i++){
      Body body=getBody(i);
      Vector3D mg=(Vector3D)g.clone();
      mg.mul(body.mass);
      Matrix m=body.massMatrix();
      Vector f=body.forceVector
      	(mg,new Vector3D(),getPosition(i,step));
      for(int i1=0;i1<6;i1++){
      	for(int j1=0;j1<6;j1++){
      	  matrix.set(ii+i1,ii+j1,m.get(i1,j1));
      	}
      	vector.set(ii+i1,f.get(i1));
      }
      ii+=6;
    }

    int constraintCount=getConstraintCount();
    for(int i=0;i<constraintCount;i++){
      RigidConstraint c=getConstraint(i);
      int dof=c.dof();
      c.create(getPosition(c.p1(),step),getPosition(c.p2(),step));
      Matrix m[]=c.getMatrix();
      Vector f=c.getVector();
      if(c.p1()>=0){
      	int j0=6*c.p1();
      	for(int i1=0;i1<dof;i1++){
      	  for(int j1=0;j1<6;j1++){
      	    matrix.set(ii+i1,j0+j1,m[0].get(i1,j1));
      	    matrix.set(j0+j1,ii+i1,m[0].get(i1,j1));
      	  }
      	}
      }
      if(c.p2()>=0){
      	int j0=6*c.p2();
      	for(int i1=0;i1<dof;i1++){
      	  for(int j1=0;j1<6;j1++){
      	    matrix.set(ii+i1,j0+j1,m[1].get(i1,j1));
      	    matrix.set(j0+j1,ii+i1,m[1].get(i1,j1));
      	  }
      	}
      }
      for(int i1=0;i1<dof;i1++){
      	vector.set(ii+i1,f.get(i1));
      }
      ii+=dof;
    }

    int springCount=getSpringCount();
    for(int i=0;i<springCount;i++){
      Constraint s=getSpring(i);
      int dof=s.dof();
      s.create(getPosition(s.p1(),step),getPosition(s.p2(),step));
      Vector f=s.getVector();
      if(s.p1()>=0){
      	int i0=6*s.p1();
      	for(int i1=0;i1<dof;i1++){
      	  vector.add(i0+i1,f.get(i1));
      	}
      }
      if(s.p2()>=0){
      	int i0=6*s.p2();
      	for(int i1=0;i1<dof;i1++){
      	  vector.add(i0+i1,f.get(dof+i1));
      	}
      }
    }
  }

//--------------------------------------------------------------------//
/**
 * Runge-Kutta @Xebṽf[^Ԃ
 * @param index ̂̃CfbNX
 * @param step Runge - Kutta @̃Xebv
 * @return w肳ꂽ Runge-Kutta @Xebṽf[^
 */
  protected Position getPosition(int index,int step){
    if(index<0) return null;
    return tempPos[index][step];
  }

//--------------------------------------------------------------------//
/**
 * xݒ肷
 * @param a xxNg
 * @param step Runge - Kutta @̃Xebv
 */
  protected void setAcc(Vector a,int step){
    int bodyCount=getBodyCount();
    int ii=0;
    for(int i=0;i<bodyCount;i++){
      tempPos[i][step].setAcc(a,ii);
      ii+=6;
    }
  }

//--------------------------------------------------------------------//
/**
 * f[^XV
 * @param dt ԍݕ
 */
  protected void update(double dt){
    int bodyCount=getBodyCount();
    Vector3D dx=new Vector3D();
    Vector3D dv=new Vector3D();
    Vector dp=new Vector(4);
    Vector3D dw=new Vector3D();
    double cdt[]={dt*C1[0],dt*C1[1],dt*C1[2],dt*C1[3]};
    for(int i=0;i<bodyCount;i++){
      dx.set(0.0,0.0,0.0);
      dv.set(0.0,0.0,0.0);
      dp.clear();
      dw.set(0.0,0.0,0.0);
      for(int j=0;j<4;j++){
      	addVector(dx,tempPos[i][j].v,cdt[j]);
      	addVector(dv,tempPos[i][j].a,cdt[j]);
      	dp.add(0.5*cdt[j],tempPos[i][j].eulerVel());
      	addVector(dw,tempPos[i][j].dw,cdt[j]);
      }
      getBody(i).position().update(dx,dv,dp,dw);
    }
  }

//--------------------------------------------------------------------//
/**
 * xNgɒl
 * @param v1 xNg
 * @param v2 xNg
 * @param coef W
 */
  protected static void addVector(Tuple3D v1,Tuple3D v2,double coef){
    v1.add(coef*v2.x,coef*v2.y,coef*v2.z);
  }
}
