package mbd.component;
import java.awt.Color;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Point;
import java.text.DecimalFormat;
import figure.Cube3D;
import figure.Point3D;
import figure.Transform;
import gui.BasicViewCanvas;
import mbd.MultiBodyModel;

//--------------------------------------------------------------------//
/**
 * ̌nf[^f\LoX
 */
public class MBDViewCanvas extends BasicViewCanvas{

/** ̍ŏl */
  public static final double MIN_VALUE=1.0E-20;

/** li[z */
  protected static final DecimalFormat
    FORMS[]={new DecimalFormat("0.000E0"),
      	     new DecimalFormat("#####0.0##"),
      	     new DecimalFormat("###0.0####")};

/** ̌nf[^f */
  protected MultiBodyModel model;

/** f\̈TCY */
  protected int viewSize;

/** ʒ_ */
  protected Point centerPoint;

/** G[bZ[W */
  protected String message;

/** tHg */
  protected Font font;

/** wiF */
  protected Color bgColor;

/** OiF */
  protected Color fgColor;

/** fsړW */
  protected double transCoef;

/** f]xW */
  protected double rotateCoef;

/** fgEkW */
  protected double zoomCoef;

/** ϊs */
  protected Transform transform;

/**
 * ̌nf[^f\LoX\z
 * @param model ̌nf[^f
 * @param message G[bZ[W
 * @param font tHg
 * @param bgColor wiF
 * @param fgColor OiF
 */
  public MBDViewCanvas(MultiBodyModel model,String message,Font font,
      	      	       Color bgColor,Color fgColor){
    this.model=model;
    this.message=message;
    this.font=font;
    setColors(bgColor,fgColor);
    transform=new Transform();
  }

//--------------------------------------------------------------------//
/**
 * \Fݒ肷
 * @param bgColor wiF
 * @param fgColor OiF
 */
  protected void setColors(Color bgColor,Color fgColor){
    if(bgColor==null){
      this.bgColor=Color.WHITE;
    }
    else{
      this.bgColor=bgColor;
    }
    if(fgColor==null){
      this.fgColor=Color.BLACK;
    }
    else{
      this.fgColor=fgColor;
    }
    setBackground(bgColor);
  }

//--------------------------------------------------------------------//
/**
 * ̌nf[^fԂ
 * @return ̌nf[^f
 */
  public MultiBodyModel getModel(){
    return model;
  }

//--------------------------------------------------------------------//
/**
 * f𕽍sړ
 * @param dx ʏ <i>x</i> ړ
 * @param dy ʏ <i>y</i> ړ
 */
  public void translate(int dx,int dy){
    transform.translate(transCoef*(double)dx,-transCoef*(double)dy,0.0);
  }

//--------------------------------------------------------------------//
/**
 * f]
 * @param dx ʏ <i>x</i> ړ
 * @param dy ʏ <i>y</i> ړ
 */
  public void rotate(int dx,int dy){
    transform.rotate(rotateCoef*(double)dy,rotateCoef*(double)dx,0.0);
  }

//--------------------------------------------------------------------//
/**
 * fgEk
 * @param dx ʏ <i>x</i> ړ
 * @param dy ʏ <i>y</i> ړ
 */
  public void zoom(int dx,int dy){
    transform.zoom(Math.pow(2.0,zoomCoef*(double)(dx+dy)));
  }

//--------------------------------------------------------------------//
/**
 * _ʓɂ邩ǂʂ
 * @param x <i>x</i> W
 * @param y <i>y</i> W
 * @return _ʓɂꍇ true
 */
  protected boolean innerFrame(int x,int y){
    return ((0<=x) && (x<panelSize.width)
      	 && (0<=y) && (y<panelSize.height));
  }

//--------------------------------------------------------------------//
/**
 * `ʒu
 */
  public void initView(){
    transform.init();
    repaint();
  }

//--------------------------------------------------------------------//
/**
 * xɂ
 */
  public void viewX(){
    transform.init();
    transform.rotate(0.0,1.0,0.0,-0.5*Math.PI);
    transform.rotate(0.0,0.0,1.0,-0.5*Math.PI);
    repaint();
  }

//--------------------------------------------------------------------//
/**
 * yɂ
 */
  public void viewY(){
    transform.init();
    transform.rotate(1.0,0.0,0.0,-0.5*Math.PI);
    repaint();
  }

//--------------------------------------------------------------------//
/**
 * ĕ`悷
 * @param g `悷OtBbNXReLXg
 */
  public void update(Graphics g){
    if(model!=null){
      paintModel(bg);
    }
    else{
      drawErrorMessage(bg,message);
    }
    g.drawImage(bgImage,0,0,this);
  }

//--------------------------------------------------------------------//
/**
 * `悷
 * @param g `悷OtBbNXReLXg
 */
  public void paint(Graphics g){
    setPaint();
    if(model!=null){
      paintModel(bg);
    }
    else{
      drawErrorMessage(bg,message);
    }
    g.drawImage(bgImage,0,0,this);
  }

//--------------------------------------------------------------------//
/**
 * ʐݒ
 */
  protected void setPaint(){
    super.setPaint();
    viewSize=Math.min(panelSize.width,panelSize.height);
    centerPoint=new Point(viewSize/2,viewSize/2);
    transCoef=1.0;
    rotateCoef=Math.PI/(double)viewSize;
    zoomCoef=1.0/(double)viewSize;
    bg.setFont(font);
  }

//--------------------------------------------------------------------//
/**
 * ̂`悷
 * @param g `悷OtBbNXReLXg
 */
  protected void paintModel(Graphics2D g){
    g.setColor(bgColor);
    g.fillRect(0,0,panelSize.width,panelSize.height);

    Cube3D bounds=model.getBounds();
    Point3D center=bounds.getCenterLocation();
    double boundSize=Math.max(Math.max(bounds.width,bounds.height),
      	      	      	      bounds.depth);
    double vCoef=0.8*(double)viewSize/Math.max(boundSize,MIN_VALUE);

    g.setColor(fgColor);
    int bodyCount=model.getBodyCount();
    for(int i=0;i<bodyCount;i++){
      paintBody(g,i,vCoef,center);
    }
    paintCoord(g,0.2*(double)viewSize,center);
    g.drawString(time(model.getTime().time()),10,20);
  }

//--------------------------------------------------------------------//
/**
 * ̂`悷
 * @param g `悷OtBbNXReLXg
 * @param index ̂̃CfbNX
 * @param vCoef f̕\{
 * @param center f̒S\ <code>Point3D</code>
 */
  protected void paintBody(Graphics2D g,int index,double vCoef,
      	      	      	   Point3D center){
    Point3D pt[][]=model.getBody(index).vertices();
    int n1=pt.length;
    int n2=pt[0].length;
    Point p[][]=new Point[n1][n2];
    for(int i=0;i<n1;i++){
      for(int j=0;j<n2;j++){
      	ViewPoint3D vp=new ViewPoint3D(vCoef*(pt[i][j].x-center.x),
      	      	      	      	       vCoef*(pt[i][j].y-center.y),
      	      	      	      	       vCoef*(pt[i][j].z-center.z));
      	transform.trans(vp);
      	p[i][j]=vp.getPoint(centerPoint);
      	g.fillOval(p[i][j].x-2,p[i][j].y-2,4,4);
      }
    }
    for(int i=0;i<n1;i++){
      for(int j=0;j<n2;j++){
      	int j2=(j+1)%n2;
      	g.drawLine(p[i][j].x,p[i][j].y,p[i][j2].x,p[i][j2].y);
      }
    }
    for(int i=0;i<n1-1;i++){
      for(int j=0;j<n2;j++){
      	g.drawLine(p[i][j].x,p[i][j].y,p[i+1][j].x,p[i+1][j].y);
      }
    }
  }

//--------------------------------------------------------------------//
/**
 * W`悷
 * @param g `悷OtBbNXReLXg
 * @param size W̕`TCY
 * @param point _`ʒu
 */
  protected void paintCoord(Graphics2D g,double size,Point3D point){
    ViewPoint3D vp[]=new ViewPoint3D[4];
    vp[0]=new ViewPoint3D(0.0,0.0,0.0);
    vp[1]=new ViewPoint3D(size,0.0,0.0);
    vp[2]=new ViewPoint3D(0.0,size,0.0);
    vp[3]=new ViewPoint3D(0.0,0.0,size);
    for(int i=0;i<4;i++){
      vp[i].translate(point.x,point.y,point.z);
      transform.trans(vp[i]);
    }
    Point p0=vp[0].getPoint(centerPoint);
    String l[]={"X","Y","Z"};
    for(int i=0;i<3;i++){
      Point p1=vp[i+1].getPoint(centerPoint);
      g.drawLine(p0.x,p0.y,p1.x,p1.y);
      g.drawString(l[i],p1.x,p1.y-5);
    }
  }

//--------------------------------------------------------------------//
/**
 * ̕\Ԃ
 * @param time 
 * @return ̕\
 */
  private String time(double time){
    if(time>=1.0E5){
      return "Time = "+FORMS[0].format(time);
    }
    else if((time>=1.0) || (time==0.0)){
      return "Time = "+FORMS[1].format(time);
    }
    else if(time>=0.001){
      return "Time = "+FORMS[2].format(time);
    }
    else{
      return "Time = "+FORMS[0].format(time);
    }
  }
}
