package mbd.io;
import java.io.BufferedReader;
import figure.Point3D;
import figure.Vector3D;
import fileio.TextDataReader;
import mbd.Body;
import mbd.MultiBodyModel;
import mbd.Position;
import mbd.body.Box;
import mbd.body.Cylinder;
import mbd.body.Sphere;
import mbd.constraint.BallJoint;
import mbd.constraint.CylinderJoint;
import mbd.constraint.RotateJoint;
import mbd.constraint.Slider;
import mbd.constraint.Spring;
import mbd.constraint.UniversalJoint;

//--------------------------------------------------------------------//
/**
 * ̌nf[^fǂݍ݃NX
 */
public class MultiBodyModelReader extends TextDataReader{

/** ̌nf[^f */
  protected MultiBodyModel model;

/**
 * ̌nf[^fǂݍ݃IuWFNg\z
 * @param model ̌nf[^f
 */
  public MultiBodyModelReader(MultiBodyModel model){
    this.model=model;
  }

//--------------------------------------------------------------------//
/**
 * ̌nf[^fǂݍ
 * @param br ̓[_[
 * @throws Exception 炩̗RŃt@Cǂ߂ȂƂ
 */
  public void readData(BufferedReader br) throws Exception{
    double alpha=0.0;
    double beta=0.0;
    while(true){
      String line=br.readLine();
      if(line==null) break;
      String ss[]=split(line);
      if(ss.length>0){
      	String keyWord=ss[0].toLowerCase();
      	if(keyWord.equals("dt")){
      	  double dt=Double.parseDouble(ss[1]);
      	  model.setDt(dt);
      	}
      	else if(keyWord.equals("baumgarte")){
      	  alpha=Double.parseDouble(ss[1]);
      	  beta=Double.parseDouble(ss[2]);
      	}
      	else if(keyWord.equals("box")){
      	  Box box=readBox(ss);
      	  setPosition(box,br.readLine());
      	  model.addBody(box);
      	}
      	else if(keyWord.equals("cylinder")){
      	  Cylinder cyl=readCylinder(ss);
      	  setPosition(cyl,br.readLine());
      	  model.addBody(cyl);
      	}
      	else if(keyWord.equals("sphere")){
      	  Sphere sph=readSphere(ss);
      	  setPosition(sph,br.readLine());
      	  model.addBody(sph);
      	}
      	else if(keyWord.equals("balljoint")){
      	  model.addConstraint(readBallJoint(ss));
      	}
      	else if(keyWord.equals("universaljoint")){
      	  model.addConstraint(readUniversalJoint(ss));
      	}
      	else if(keyWord.equals("rotatejoint")){
      	  model.addConstraint(readRotateJoint(ss));
      	}
      	else if(keyWord.equals("cylinderjoint")){
      	  model.addConstraint(readCylinderJoint(ss));
      	}
      	else if(keyWord.equals("slider")){
      	  model.addConstraint(readSlider(ss));
      	}
      	else if(keyWord.equals("spring")){
      	  model.addSpring(readSpring(ss));
      	}
      	else if(keyWord.equals("gravity")){
      	  double x=Double.parseDouble(ss[1]);
      	  double y=Double.parseDouble(ss[2]);
      	  double z=Double.parseDouble(ss[3]);
      	  model.setGravity(new Vector3D(x,y,z));
      	}
      }
    }
    if((alpha>0.0) && (beta>0.0)) model.setBaumgarte(alpha,beta);
    model.init();
  }

//--------------------------------------------------------------------//
/**
 * pǂݍ
 * @param ss 
 * @return p
 * @throws Exception 炩̗RŃt@Cǂ߂ȂƂ
 */
  protected Box readBox(String ss[]) throws Exception{
    int label=Integer.parseInt(ss[1]);
    double a=Double.parseDouble(ss[2]);
    double b=Double.parseDouble(ss[3]);
    double c=Double.parseDouble(ss[4]);
    double dens=Double.parseDouble(ss[5]);
    return new Box(label,a,b,c,dens);
  }

//--------------------------------------------------------------------//
/**
 * ~ǂݍ
 * @param ss 
 * @return ~
 * @throws Exception 炩̗RŃt@Cǂ߂ȂƂ
 */
  protected Cylinder readCylinder(String ss[]) throws Exception{
    int label=Integer.parseInt(ss[1]);
    double ro=Double.parseDouble(ss[2]);
    double ri=Double.parseDouble(ss[3]);
    double l=Double.parseDouble(ss[4]);
    double dens=Double.parseDouble(ss[5]);
    return new Cylinder(label,ro,ri,l,dens);
  }

//--------------------------------------------------------------------//
/**
 * 󋅂ǂݍ
 * @param ss 
 * @return 
 * @throws Exception 炩̗RŃt@Cǂ߂ȂƂ
 */
  protected Sphere readSphere(String ss[]) throws Exception{
    int label=Integer.parseInt(ss[1]);
    double ro=Double.parseDouble(ss[2]);
    double ri=Double.parseDouble(ss[3]);
    double dens=Double.parseDouble(ss[4]);
    return new Sphere(label,ro,ri,dens);
  }

//--------------------------------------------------------------------//
/**
 * ̂̎pݒ肷
 * @param body 
 * @param line 
 * @throws Exception 炩̗RŃt@Cǂ߂ȂƂ
 */
  protected void setPosition(Body body,String line) throws Exception{
    String ss[]=split(line);
    String keyWord=ss[0].toLowerCase();
    if(keyWord.equals("euler")){
      double phy=Math.toRadians(Double.parseDouble(ss[1]));
      double th=Math.toRadians(Double.parseDouble(ss[2]));
      double psi=Math.toRadians(Double.parseDouble(ss[3]));
      double dphy=Math.toRadians(Double.parseDouble(ss[4]));
      double dth=Math.toRadians(Double.parseDouble(ss[5]));
      double dpsi=Math.toRadians(Double.parseDouble(ss[6]));
      setPos(body.position(),phy,th,psi,dphy,dth,dpsi);
    }
  }

//--------------------------------------------------------------------//
/**
 * IC[pʒuݒ肷
 * @param p ʒu
 * @param phy IC[p
 * @param th IC[p
 * @param psi IC[p
 * @param dphy IC[pӂ̎Ԕ
 * @param dth IC[pƂ̎Ԕ
 * @param dpsi IC[pՂ̎Ԕ
 */
  protected void setPos(Position p,double phy,double th,double psi,
      	      	      	double dphy,double dth,double dpsi){
    double sphy=Math.sin(phy);
    double cphy=Math.cos(phy);
    double sth=Math.sin(th);
    double cth=Math.cos(th);
    double spsi=Math.sin(psi);
    double cpsi=Math.cos(psi);
    double p0=0.5*Math.sqrt((1.0+Math.cos(phy+psi))*(1.0+cth));
    if(p0>0.0){
      double p1=0.25*sth*(cpsi+cphy)/p0;
      double p2=0.25*sth*(sphy-spsi)/p0;
      double p3=0.25*(1.0+cth)*Math.sin(phy+psi)/p0;
      p.p.set(p0,p1,p2,p3);
    }
    else{
      double p1=Math.sqrt(0.5*(cphy*cpsi-sphy*cth*spsi+1.0));
      if(p1>0.0){
      	double p2=-0.25*Math.sin(phy-psi)*(1.0-cth)/p1;
      	double p3=0.25*sth*(sphy+spsi)/p1;
      	p.p.set(p0,p1,p2,p3);
      }
      else{
      	double p2=Math.sqrt(0.5*(-sphy*spsi+cphy*cth*cpsi+1.0));
      	if(p2>0.0){
      	  double p3=-0.25*sth*(cphy-cpsi)/p2;
      	  p.p.set(p0,p1,p2,p3);
      	}
      	else{
      	  double p3=Math.sqrt(0.5*(cth+1.0));
      	  p.p.set(p0,p1,p2,p3);
      	}
      }
    }
    p.w.set(sth*spsi*dphy+cpsi*dth,sth*cpsi*dphy-spsi*dth,
      	    cth*spsi*dphy+dpsi);
  }

//--------------------------------------------------------------------//
/**
 * WCgǂݍ
 * @param ss 
 * @return WCg
 * @throws Exception 炩̗RŃt@Cǂ߂ȂƂ
 */
  protected BallJoint readBallJoint(String ss[]) throws Exception{
    int p1=Integer.parseInt(ss[1]);
    int p2=Integer.parseInt(ss[2]);
    double x1=Double.parseDouble(ss[3]);
    double y1=Double.parseDouble(ss[4]);
    double z1=Double.parseDouble(ss[5]);
    double x2=Double.parseDouble(ss[6]);
    double y2=Double.parseDouble(ss[7]);
    double z2=Double.parseDouble(ss[8]);
    return new BallJoint(p1,p2,new Point3D(x1,y1,z1),
      	      	      	 new Point3D(x2,y2,z2));
  }

//--------------------------------------------------------------------//
/**
 * jo[TWCgǂݍ
 * @param ss 
 * @return jo[TWCg
 * @throws Exception 炩̗RŃt@Cǂ߂ȂƂ
 */
  protected UniversalJoint readUniversalJoint(String ss[])
    throws Exception{

    int p1=Integer.parseInt(ss[1]);
    int p2=Integer.parseInt(ss[2]);
    double x1=Double.parseDouble(ss[3]);
    double y1=Double.parseDouble(ss[4]);
    double z1=Double.parseDouble(ss[5]);
    double x2=Double.parseDouble(ss[6]);
    double y2=Double.parseDouble(ss[7]);
    double z2=Double.parseDouble(ss[8]);
    return new UniversalJoint(p1,p2,new Point3D(x1,y1,z1),
      	      	      	      new Point3D(x2,y2,z2));
  }

//--------------------------------------------------------------------//
/**
 * ]WCgǂݍ
 * @param ss 
 * @return ]WCg
 * @throws Exception 炩̗RŃt@Cǂ߂ȂƂ
 */
  protected RotateJoint readRotateJoint(String ss[]) throws Exception{
    int p1=Integer.parseInt(ss[1]);
    int p2=Integer.parseInt(ss[2]);
    double x1=Double.parseDouble(ss[3]);
    double y1=Double.parseDouble(ss[4]);
    double z1=Double.parseDouble(ss[5]);
    double x2=Double.parseDouble(ss[6]);
    double y2=Double.parseDouble(ss[7]);
    double z2=Double.parseDouble(ss[8]);
    return new RotateJoint(p1,p2,new Point3D(x1,y1,z1),
      	      	      	   new Point3D(x2,y2,z2));
  }

//--------------------------------------------------------------------//
/**
 * ]~WCgǂݍ
 * @param ss 
 * @return ]~WCg
 * @throws Exception 炩̗RŃt@Cǂ߂ȂƂ
 */
  protected CylinderJoint readCylinderJoint(String ss[])
    throws Exception{

    int p1=Integer.parseInt(ss[1]);
    int p2=Integer.parseInt(ss[2]);
    double x1=Double.parseDouble(ss[3]);
    double y1=Double.parseDouble(ss[4]);
    double z1=Double.parseDouble(ss[5]);
    double x2=Double.parseDouble(ss[6]);
    double y2=Double.parseDouble(ss[7]);
    double z2=Double.parseDouble(ss[8]);
    double z0=Double.parseDouble(ss[9]);
    return new CylinderJoint(p1,p2,new Point3D(x1,y1,z1),
      	      	      new Point3D(x2,y2,z2),z0);
  }

//--------------------------------------------------------------------//
/**
 * ]iWCgǂݍ
 * @param ss 
 * @return ]iWCg
 * @throws Exception 炩̗RŃt@Cǂ߂ȂƂ
 */
  protected Slider readSlider(String ss[]) throws Exception{
    int p1=Integer.parseInt(ss[1]);
    int p2=Integer.parseInt(ss[2]);
    double x1=Double.parseDouble(ss[3]);
    double y1=Double.parseDouble(ss[4]);
    double z1=Double.parseDouble(ss[5]);
    double x2=Double.parseDouble(ss[6]);
    double y2=Double.parseDouble(ss[7]);
    double z2=Double.parseDouble(ss[8]);
    double z0=Double.parseDouble(ss[9]);
    return new Slider(p1,p2,new Point3D(x1,y1,z1),
      	      	      new Point3D(x2,y2,z2),z0);
  }

//--------------------------------------------------------------------//
/**
 * ΂˗vfǂݍ
 * @param ss 
 * @return ΂˗vf
 * @throws Exception 炩̗RŃt@Cǂ߂ȂƂ
 */
  protected Spring readSpring(String ss[]) throws Exception{
    int p1=Integer.parseInt(ss[1]);
    int p2=Integer.parseInt(ss[2]);
    double k=Double.parseDouble(ss[3]);
    double l0=Double.parseDouble(ss[4]);
    double x1=Double.parseDouble(ss[5]);
    double y1=Double.parseDouble(ss[6]);
    double z1=Double.parseDouble(ss[7]);
    double x2=Double.parseDouble(ss[8]);
    double y2=Double.parseDouble(ss[9]);
    double z2=Double.parseDouble(ss[10]);
    return new Spring(p1,p2,new Point3D(x1,y1,z1),
      	      	      new Point3D(x2,y2,z2),k,l0);
  }
}
